// Action.cc -*- c++ -*- Hans-Peter Stoerr Thu Feb  4 1999

#include <string>
#include <vector>
#include <map>

#include "general.h"
#include "pddldomain.h"
#include "Domain.h"
#include "Action.h"
#include "misc/mystl.h"

ActionIterator::ActionIterator()
  : mpDom(NULL)
{
  mAction.mName="";             // sign for end()
}

ActionIterator::ActionIterator(Domain& theDom)
  : mpDom(&theDom)
{
  mActIter=mpDom->dom.actions.begin();
  mActEnd=mpDom->dom.actions.end();
  NextAction(TRUE);
}

ActionIterator::~ActionIterator()
{
  NOOP;
}

bool    ActionIterator::operator==(const ActionIterator& b) const
{
  if (""==mAction.mName && ""==b.mAction.mName) return TRUE;
  return (b.mAction==mAction);
}

bool    ActionIterator::operator!=(const ActionIterator& b) const
{
  return !(*this == b);
}

bool    ActionIterator::NextAction(bool firsttime)
{
  assert(mActIter!=mActEnd);
  mAction.mName="";
  if (!firsttime) mActIter++;
  if( mActIter==mActEnd ) return FALSE; // we are done.
  mAction.mName=mActIter->second.name;
  mAction.mpRep=&mActIter->second;
  // now we initialize the iterator vectors
  mBeginIters=vector<StringVector::const_iterator>();
  mIters=mBeginIters;
  mEndIters=mBeginIters;
  mVarNames=StringVector();
  mAction.mSubst.reset();
  StringStringPairList::const_iterator it;
  for(it=mActIter->second.parameters.begin();
      mActIter->second.parameters.end()!=it; ++it) {
    mBeginIters.push_back(mpDom->mTypeConstantsMap[it->second].begin());
    mIters.push_back(mpDom->mTypeConstantsMap[it->second].begin());
    mEndIters.push_back(mpDom->mTypeConstantsMap[it->second].end());
    mAction.mSubst[it->first]=*mIters.back();
    mVarNames.push_back(it->first);
  }
  return TRUE;
}

ActionIterator& ActionIterator::operator++()
{
  assert(""!=mAction.mName);           // we can't increment enditerator.
  if (""==mAction.mName || 0==mIters.size()) { // end reached.
    NextAction();
    return *this;
  }
  vector<StringVector::const_iterator>::const_iterator itb; // iter on mBeginIters
  vector<StringVector::const_iterator>::iterator iti; // iter on mIters
  vector<StringVector::const_iterator>::const_iterator ite; // iter on mEndIters
  StringVector::iterator ita;   // iterator on mArgs
  itb=mBeginIters.end()-1, iti=mIters.end()-1, ite=mEndIters.end()-1;
  ita=mVarNames.end()-1;
  ++(*iti);
  while ( *iti==*ite ) {        // while the corresponding arg is finished
    if ( mIters.begin()==iti ) { // first arg went through, so we are ready with mAction.mName
      if (NextAction()) return *this; // if there is a next fluent we return this
      mAction.mName=""; return *this;  // we are done with all fluents.
    }
    *iti=*itb;                  // reset iterator
    mAction.mSubst[*ita]=**iti; // set arg to first object of the corresponding type
    --itb; --iti; --ite; --ita; // goto next argument
    ++(*iti);
  }
  mAction.mSubst[*ita]=**iti;
  return *this;
}

// ================================================================
// Action Implementation

void    Action::Printit(ostream& s) const
{
  s << "Action: " << mName << ", subst: " << mSubst << "\n";
}

string  Action::Description() const
{
  string res(mName+"(");
  FirsttimeNexttime<string> sep("",",");
  StringStringPairList::const_iterator it=mpRep->parameters.begin();
  for( ; it!=mpRep->parameters.end(); ++it )
    res+=sep()+mSubst(it->first);
  return res+")";
}
