// BddTransPlan.cc -*- c++ -*- Hans-Peter Stoerr Wed Feb  3 1999

#include <iostream.h>
#include <algo.h>
#include <iterator.h>
#include "Action.h"
#include "BddTransPlan.h"
#include "misc/mystl.h"

typedef  set<Bdd,Bdd::BddCompare> BddSet;
typedef  map<Bdd,BddSet,Bdd::BddCompare> BddBddSetMap;

Option noop("noop","Insert NOOP-operation into transition relation.");

Option reduce("reduce","Reduce state BDD wrt. previous state each forward step.");
Option reduce2("reduce2","Reduce state BDD wrt. to next to previous state each forward step.");
Option reduceall("reduceall","Reduce state BDD wrt. all previous states each step.");

Option reducesum("reducesum","Reduce state BDD wrt. sum of all previous states each step.");
Option betweensum("betweensum","Reduce state BDD wrt. sum of all previous states each step with Between.");

Option dontcarered("dontcarered","Reduce state BDD wrt. states which lead to nothing.");
Option dontcarebet("dontcarebet","Reduce state BDD wrt. states which lead to nothing with Between.");

Option staticfluents("staticfluents","Print number of static fluents in last state.");
static DependentOption printstaticfluents("printstaticfluents","staticfluents","Print static fluents in last state.");
Option unchangable("unchangable","Print number of fluents unchangable by transrel.");
static DependentOption printunchangable("printunchangable","unchangable","Print fluents unchangable by transrel.");
Option tryreduction("tryreduction","Print data for evaluation of T-reduction");

Option reduceactionset("reduceactionset","Compute all actions that might be relevant.");

Option nogoal("nogoal","Ignore goal (i.e. run until fixpoint for test purposes)");
Option maxlen("maxlen","Maximal plan length",400);
Option backwards("backwards","Plan backwards");
Option bidirectional("bidirectional","Plan backwards and forwards");

Option reorder("reorder","Try reordering");

static inline bool IsPrimedVar(Bdd B) {
  if (B.IsVariable())return *(B.VarName().end()-1)=='`';
  if (!B.IsVariable())return *((-B).VarName().end()-1)=='`';
}
static inline Bdd GetPrimedVar(Bdd B) {
  if (B.IsVariable()) return mgr.Var(B.VarName()+"`");
  if (!B.IsVariable()) return -mgr.Var((-B).VarName()+"`");
}
static inline Bdd UnPrimeVar(Bdd B) {
  if (B.IsVariable()) {
    string vnam=B.VarName();
    return mgr.Var(vnam.substr(0,vnam.size()-1));
  } else {
    string vnam=(-B).VarName();
    return -mgr.Var(vnam.substr(0,vnam.size()-1));
  }
}

static Bdd MinSize(Bdd a, Bdd b) // returns the Bdd with less size.
{
  return (a.Size()<b.Size()) ? a : b;
}

// static Bdd MinSize(Bdd a, Bdd b, Bdd c) // returns the Bdd with least size.
// {
//   b=MinSize(b,c);
//   return MinSize(a,b);
// }

Bdd   BddTransPlan::PrimedToUnprimed(Bdd b) const { // faster replacement for Substitute(mgr.Prim...)
  return b.RelProd(mgr.Primed(),mDom.mIdent);
}

Bdd   BddTransPlan::UnprimedToPrimed(Bdd b) const { // faster replacement for Substitute(mgr.Unprim...)
  return b.RelProd(mgr.Unprimed(),mDom.mIdent);
}

void BddTransPlan::Init()
{
  mInit=mDom.mInit;
  mGoal=mDom.mGoal;
  mFilter=mDom.mFilter;
  if(nogoal) mGoal=mgr.Zero();
  if (reorder && 10!=reorder) mgr.SetDynReordering(reorder);
  GenerateTransRel();
}

void BddTransPlan::GenerateTransRel()
{
  ActionIterator ait(mDom),aend;
  for(; aend!=ait; ++ait){
    pair<Bdd,BddVector> aBddInfo(mDom.ActionBdd(*ait));
    if(mgr.Zero()!=aBddInfo.first) {      // if action might be executable somewhere
      mActionBdds[ait->Description()]=aBddInfo;
    }
  }
  CHECK(mActionBdds.size()>0,"Huh? No executable action in *any* state?");
  if(noop) {
    pair<Bdd,BddVector> noopInfo(mDom.mIdent,BddVector());
    mActionBdds["noop"]=noopInfo;
  }
  if (reduceactionset) ReduceActionSet();
  map<string,pair<Bdd,BddVector> >::const_iterator it=mActionBdds.begin(),end=mActionBdds.end();
  for (;end!=it;++it) mTrans+=it->second;
  if (mgr.One()!=mFilter) {
    Bdd transFilter=mFilter*mFilter.Substitute(mgr.UnprimedToPrimed());
    mTrans.Finish(transFilter);
  }
  else mTrans.Finish(mgr.One());

  if (unchangable) {
    BddSet        effFlu;
    {map<string,pair<Bdd,BddVector> >::const_iterator it=mActionBdds.begin(),end=mActionBdds.end();
    for (;end!=it;++it) {
    BddVector::const_iterator itn=it->second.second.begin(), endn=it->second.second.end();
    for (; endn!=itn; ++itn) effFlu.insert(*itn); }}

    StringVector::const_iterator it=mDom.mFluents.begin(), end=mDom.mFluents.end();
    BddVector   uchgFlu;
    Bdd uchgVal=mgr.One();
    int numunchangable=0, numfluents=0;
    if (printunchangable) *gpLog << "Unchangable Fluents: ";
    for (; end!=it; ++it) {
      numfluents++;
      Bdd v(mgr.Var(*it)),vp(mgr.Var(*it+"`"));
      if (!mDom.IsConstantFluent(*it) && 
	  ( (0==effFlu.count(vp) && 0==effFlu.count(-vp)) ||
          !mTrans.Intersects(v*-vp + -v*vp))) {
        // assert(!mTrans.Intersects(v*-vp + -v*vp));
        if (printunchangable) *gpLog << " " << *it;
        uchgFlu.push_back(v);
        uchgFlu.push_back(vp);
        if (mgr.Zero()==v*mInit) uchgVal*= -v*-vp; // fluent is false
        else uchgVal*= v*vp;
        numunchangable++;
      }
    }
    if (printunchangable) *gpLog << "\n";
    *gpLog << "Number of unchangable fluents " << numunchangable/numfluents << "%: "
           << numunchangable << " out of " << numfluents << ".\n";
    // now reduce everything wrt. unchgFlu
    if (0<numunchangable) {      
      BddAssoc uchgAss(uchgFlu);
      mTrans.Reduce(uchgVal,uchgAss);
      mInit=(uchgVal*mInit).Exists(uchgAss);
      mFilter=(uchgVal*mFilter).Exists(uchgAss);
      mGoal=(uchgVal*mGoal).Exists(uchgAss);
      map<string,pair<Bdd,BddVector> >::iterator it=mActionBdds.begin(),end=mActionBdds.end();
      for (;end!=it;++it) it->second.first=(it->second.first*uchgVal).Exists(uchgAss);
    }
  }
  
  if(9<reorder) {
    BddAssoc ass;
    Bdd alltrans=mTrans.RelProd(ass,mgr.One());
    printit(alltrans);
    Timing(mgr.Reorder(CAL_REORDER_SIFT));
    printit(alltrans);  
    Timing(mgr.Reorder(CAL_REORDER_WINDOW));
    printit(alltrans);  
    Timing(mgr.Reorder(CAL_REORDER_SIFT));
    printit(alltrans);  
    Timing(mgr.Reorder(CAL_REORDER_WINDOW));
    printit(alltrans);  
    Timing(mgr.Reorder(CAL_REORDER_SIFT));
    printit(alltrans);  
    Timing(mgr.Reorder(CAL_REORDER_WINDOW));
    printit(alltrans);  
    Timing(mgr.Reorder(CAL_REORDER_SIFT));
    printit(alltrans);  
    Timing(mgr.Reorder(CAL_REORDER_WINDOW));
    printit(alltrans);  
    Timing(mgr.Reorder(CAL_REORDER_SIFT));
    printit(alltrans);  
    Timing(mgr.Reorder(CAL_REORDER_WINDOW));
    printit(alltrans);  
    Timing(mgr.Reorder(CAL_REORDER_SIFT));
    printit(alltrans);
    Timing(mgr.Reorder(CAL_REORDER_WINDOW));
    printit(alltrans);
  }
}

void BddTransPlan::ReduceActionSet()
{
  // iteratively get all actions that are relevant to solve the problem.
  BddBddSetMap       dependsOn;
  map<string,pair<Bdd,BddVector> >::const_iterator it=mActionBdds.begin(), end=mActionBdds.end();
  for (; end!=it; ++it) { // for all actions
    const BddVector&   effs=it->second.second; // the depends set of the action
    BddVector::const_iterator it1=effs.begin(),it2=effs.begin(), end=effs.end();
    for (; end!=it1; ++it1)
      if (IsPrimedVar(*it1)) {
        BddSet& dependers=dependsOn[*it1];
        for (it2=effs.begin(); end!=it2; ++it2)
          if ( !IsPrimedVar(*it2) )
            dependers.insert(*it2);
      }
  }
  BddSet        oldS,workS,newS;
  // put all goals into newS.
  {BddVector     gsup=mGoal.Support();
  insert_iterator<BddSet> ins(newS,newS.end());
  copy(gsup.begin(),gsup.end(),ins);}
  // now get transitive closure
  while( 0!=newS.size() ) {
    workS=newS;
    { insert_iterator<BddSet> ins(oldS,oldS.end()); copy(workS.begin(),workS.end(),ins); }
    newS=BddSet();
    BddSet::const_iterator it=workS.begin(), end=workS.end();
    for (;end!=it; ++it) {
      const BddSet& dependers=dependsOn[GetPrimedVar(*it)];
      BddSet::const_iterator it1=dependers.begin(), end1=dependers.end();
      for (; end1!=it1; ++it1)
        if ( 0==oldS.count(*it1) ) newS.insert(*it1);
    }
  }
  // remove all actions which don't deliver any effect of oldS.
  Bdd changeOldS=mgr.Zero(); // will contain any of oldS changes
  { BddSet::const_iterator it=oldS.begin(), end=oldS.end();
  for (; end!=it; ++it) changeOldS+= *it^GetPrimedVar(*it); }
  // now filter actions.
  int countactions=0;
  map<string,pair<Bdd,BddVector> > newActionBdds;
  {
    map<string,pair<Bdd,BddVector> >::const_iterator it=mActionBdds.begin(), end=mActionBdds.end();
    for (; end!=it; ++it)
      if (mgr.Zero()!=changeOldS*it->second.first) // action doesn't change oldS
        newActionBdds[it->first]=it->second;
      else { ++countactions; }
  }
  mActionBdds=newActionBdds;
  *gpLog << "ReduceActionSet: " << changeOldS.Support().size()/2
         << " req. effect fluents, removed " << countactions << " actions .\n" << flush;
}

void BddTransPlan::Printit(ostream& s) const
{
  s << "TransPlan\n";
  s << " mInit:\t" << mInit << "\n";
  s << " mTrans:\t" << mTrans.RelProd(BddVector(),mgr.One()) << "\n";
  s << " mFilter:\t" << mFilter << "\n";
  s << " mGoal:\t" << mGoal << "\n";
  s << " mActionBdds:\t" << mActionBdds << "\n";
  s << ".\n";
}

void BddTransPlan::ForwardPass()
{
  forward.push_back(mInit);
  backward.push_back(mGoal);
  
  Bdd   statesum=mgr.Zero();

  Bdd   dontcare=mgr.Zero();
  if(dontcarered || dontcarebet) {
    dontcare=-mTrans.Exists(mgr.Primed());
    dontcare.PrintInfo("dontcare");
    backward.back()=backward.back()*-dontcare;
  }

  double last_forward_time=0;   // time taken for last forward step
  double last_backward_time=0;  // time taken for last backward step
  int   number_backward_overflows=0; // number of overflows we had with backwards steps
  long  last_backward_size=0;   // size of last BDD in forward
  long  last_forward_size=0;    // size of last BDD in backward
  long  last_backward_limit=0;  // nodelimit for last backwards failure.

  *gpLog << "First pass: [";
  { FirsttimeNexttime<string> sep("",",");
  if (!backwards||bidirectional) *gpLog << sep() << forward.back().Size() << flush;
  if (backwards||bidirectional) *gpLog << sep() << -backward.back().Size() << flush; 
  }
  while(mgr.Zero()==forward.back()*backward.back() && // goal reached
        unsigned(maxlen)>forward.size()+backward.size() && // max length of plan reached
        mgr.Zero()!=forward.back() && // nirvana reached
        (nogoal || mgr.Zero()!=backward.back()) && 
        (2>forward.size() || forward.back()!=*(forward.end()-2)) && // cycles
        (3>forward.size() || forward.back()!=*(forward.end()-3)) &&
        (2>backward.size() || backward.back()!=*(backward.end()-2)) &&
        (3>backward.size() || backward.back()!=*(backward.end()-3))) {
    
//     { % test execute all actions.
//       map<string,pair<Bdd,BddVector> >::const_iterator ait(mActionBdds.begin()),aend(mActionBdds.end());
//       for(; aend!=ait; ++ait) if (mgr.Zero!=forward.back()*ait->second)
//         cout << "Result " << ait->first << ": " << (forward.back()*ait->second).Exists(mgr.Unprimed()) << "\n";
//     } 
    
    bool go_forward=TRUE;       // "do we go forward or backward?"
    long backward_limit=3*max(last_forward_size,last_backward_size);
    if (backwards) go_forward=FALSE;
    if (bidirectional && !nogoal) {
      if ( last_forward_time <= last_backward_time ||
           3 <= number_backward_overflows || backward_limit < last_backward_limit + 10000 ||
           last_forward_size < last_backward_size ) go_forward=TRUE;
      else go_forward=FALSE;
    }
    
    Bdd nextstate;
    if (!go_forward) {
      // backward steps are prone to big BDDs and large calculation times, so we try to prevent this.
      long oldlim=0;
      if (!backwards) oldlim=mgr.SetNodeLimit(mgr.GetNumNodes()+1000+backward_limit);
      else last_backward_time=1e99; // remove all limits if we are going backwards anyway.
      Timing( GetTiming(last_backward_time,
                        nextstate=mTrans.RelProd(mgr.Primed(),UnprimedToPrimed(backward.back())
                                                 ,max(last_forward_time,last_backward_time)*3) ) );
      if (mgr.WasOverflow()) {
        assert(!backwards);
        ++number_backward_overflows;
        go_forward=TRUE;
      }
      else {
        last_backward_size=nextstate.Size();
        last_backward_limit=0;
      }
      if (!backwards) mgr.SetNodeLimit(oldlim); // reset original limit.
    }
    assert(!backwards||!go_forward);
    if(go_forward) {
      Timing( GetTiming(last_forward_time,
                        nextstate=PrimedToUnprimed(
                        mTrans.RelProd(mgr.Unprimed(),forward.back())) ) );
      last_forward_size=nextstate.Size();
    }
    if (mgr.One()!=mFilter) nextstate*=mFilter;

    // now we do some reductions. Only if requested, of course.
    if(reduce) {
      if (go_forward) Timing(nextstate=nextstate.Reduce(-forward.back()););
      else Timing(nextstate=nextstate.Reduce(-backward.back()););
    }
    if(reduceall && go_forward) { // I'm too lazy to implement this for backwards
      BddVector::const_iterator it=forward.begin();
      for(; forward.end()!=it; ++it)
        nextstate=nextstate.Reduce(-*it);
    }
    if(reduce2) {
      if (go_forward) {
        if (2<=forward.size())
          Timing(nextstate=nextstate.Reduce(-*(forward.end()-2)););
      } else { // backward
        if (2<=backward.size())
          Timing(nextstate=nextstate.Reduce(-*(forward.end()-2)););
      }
    }
    if(reducesum && go_forward) { // I'm too lazy to implement this for backwards
      Timing(nextstate=nextstate.Reduce(-statesum));
      Timing(statesum+=nextstate);
    }
    if (betweensum && go_forward ) { // I'm too lazy to implement this for backwards
      Timing(nextstate=Between(nextstate*-statesum, nextstate+statesum));
      Timing(statesum+=nextstate);
    }
    if (dontcarebet && go_forward ) { // I'm too lazy to implement this for backwards
      Timing(nextstate=Between(nextstate*-(dontcare+statesum), nextstate+statesum+dontcare));
      Timing(statesum+=nextstate);
    }
    if (dontcarered && go_forward) { // I'm too lazy to implement this for backwards
      Timing(nextstate=MinSize(nextstate,nextstate.Reduce(-(statesum+dontcare))));
      Timing(statesum+=nextstate);
    }
    
//     if(2<=states.size() && go_forward) {
//       *gpLog << "(" << nextstate.Reduce(-statesum).Size() << "/"
//              << Between(nextstate*-statesum, nextstate+statesum).Size() << "/"
//              << Between(nextstate*-dontcare, nextstate+dontcare) << "/"
//              << nextstate.Reduce(-*(states.end()-2)).Size()
//              << ")";
//       statesum+=nextstate;
//     }
    
//     {
//       printit(statesum );
//       printit(nextstate);
//       printit(nextstate*-statesum);
//       printit(nextstate.Reduce(-statesum));
//       printit(nextstate.Constrain(-statesum));
//       printit(Between(nextstate*-statesum,nextstate+statesum));
//       statesum+=nextstate;
//     }

    // now we assess T-Reduction, if requested.
    if (tryreduction && forward.size()>0 && go_forward) {
      BddVector notChanged(nextstate.SuperCube().Support());
      BddVector nChngPP(notChanged);
      BddVector::const_iterator it;
      for (it=notChanged.begin(); notChanged.end()!=it; ++it) {
        nChngPP.push_back(it->VarName()+"`");
      }
      *gpLog << "("
             << notChanged.size() << ","
             << forward.back().Exists(notChanged).Size() << flush
             << "," << mTrans.RelProd(nChngPP,mgr.One()).Size() << flush;
      map<string,pair<Bdd,BddVector> >::const_iterator it_a;
      Bdd       transn(mgr.Zero()),statecube(forward.back().SuperCube());
      for (it_a=mActionBdds.begin(); mActionBdds.end()!=it_a; ++it_a)
        if (mgr.Zero()!=statecube*it_a->second.first)
          transn += it_a->second.first.Exists(nChngPP);
      *gpLog << "," << transn.Size() << flush;
      transn=mgr.Zero();
      for (it_a=mActionBdds.begin(); mActionBdds.end()!=it_a; ++it_a)
        if (mgr.Zero()!=statecube*it_a->second.first)
          transn += it_a->second.first;
      *gpLog << "," << transn.Size() << ")" << flush;
    }

    // reductions done, state is inserted.
    if (go_forward) forward.push_back(nextstate);
    else backward.push_back(nextstate);

    if (go_forward) *gpLog << "," << nextstate.Size() << flush;
    else *gpLog << "," << -nextstate.Size() << flush;
  }
  
  *gpLog << "] nodes.\n";

  // now some statistics.
  // we check how many of our fluents havent changed yet. (In forward pass.)
  if (staticfluents) {
    Bdd changes=mInit^accumulate(forward.begin(), forward.end(), mgr.Zero());
    changes=changes.SuperCube();
    StringVector::const_iterator it=mDom.mFluents.begin(), end=mDom.mFluents.end();
    int numstaticfluents=0, numfluents=0;
    if (printstaticfluents) *gpLog << "Static Fluents: ";
    for (; end!=it; ++it) {
      numfluents++;
      if (!mDom.IsConstantFluent(*it) &&
          (mgr.Zero()==mgr.Var(*it)*changes || mgr.Zero()==(-mgr.Var(*it))*changes) ) {
        if (printstaticfluents) *gpLog << " " << *it;
        numstaticfluents++;
      }
    }
    if (printstaticfluents) *gpLog << "\n";
    *gpLog << "Number of static fluents " << numstaticfluents << " out of " << numfluents
           << " " << int(double(numstaticfluents)/numfluents) << "%.\n";
  }
}

void BddTransPlan::BackwardPass() 
{
  mPlan=StringVector();
  if (unsigned(maxlen)<=forward.size()+backward.size() && mgr.Zero()==forward.back()*backward.back() ) {
    *gpLog << "Plan: " << mDom.prob.name << ": There is likely no plan (maxlen reached).\n";
    return;
  }
  
  // have we reached a fixpoint and no plan?
  if ( mgr.Zero()==forward.back()*backward.back() ) { // if there is no plan
    if ( ( 2<=forward.size() && forward.back()==*(forward.end()-2) ) ||
         ( 3<=forward.size() && forward.back()==*(forward.end()-3) ) )
    { // and we have reached a fixpoint forwards
      *gpLog << "Plan: " << mDom.prob.name << ": There is no plan (fixpoint reached).\n";
      return;
    }
    if ( ( 2<=backward.size() && backward.back()==*(backward.end()-2) ) ||
         ( 3<=backward.size() && backward.back()==*(backward.end()-3) ) )
    {                           // and we reached a fixpoint backwards
      *gpLog << "Plan: " << mDom.prob.name << ": There is no plan (fixpoint reached).\n";
      return;
    }
  }

  if (mgr.Zero()==forward.back()) {
    *gpLog << "Plan: " << mDom.prob.name << ": There is no plan (empty stateset reached).\n";
      return;
    }
  if (mgr.Zero()==backward.back()) {
    *gpLog << "Plan: " << mDom.prob.name << ": There is no plan (empty stateset reached backwards).\n";
      return;
    }  

  // now we extract a plan.
  // first the forward part.
  forward.back()=(backward.back()*forward.back()).SatisfySupport(mgr.Unprimed());
  assert(mgr.Zero()!=forward.back());
  
  BddVector::iterator bitr=forward.end();
  --bitr;
  while (forward.begin() != bitr){
    --bitr;
    *bitr=mTrans.RelProd(mgr.Primed(),(*bitr)*UnprimedToPrimed(*(bitr+1)))
      .SatisfySupport(mgr.Unprimed());
  }
  // now the backward part.
  backward.back()=forward.back();
  bitr=backward.end();
  --bitr;
  while (backward.begin() != bitr){
    --bitr;
    *bitr=PrimedToUnprimed(mTrans.RelProd(mgr.Unprimed(),*(bitr+1)*UnprimedToPrimed(*bitr))
                           .SatisfySupport(mgr.Primed()));
  }
  // append backward part reversed to forward.
  backward.pop_back();
  while (0<backward.size()) { forward.push_back(backward.back()); backward.pop_back(); }
  // now we have to find the actions
  *gpLog << "Plan: " << mDom.prob.name << ": len " << forward.size()-1 << ": ";
  FirsttimeNexttime<string> sep("","; ");
  // :TODO: check if this works for a plan with one action!
  for(bitr=forward.begin(); forward.end()!=bitr+1; ++bitr){
    Bdd before=*bitr;
    Bdd after=*(bitr+1);
    set<string> possibleActions;
    if (0) {
      Bdd both=(before+after).SuperCube();
      {
        Bdd a=before+after;
        Timing(a.SuperCube());
        Timing(a.Support());
        Timing(a.Size());
      }
      BddVector bv(before.Support()),bov(both.Support()),difv;
      sort(bv.begin(),bv.end(),Bdd::BddCompare());
      sort(bov.begin(),bov.end(),Bdd::BddCompare());
      set_difference(bv.begin(),bv.end(),bov.begin(),bov.end(),
                     back_inserter(difv),Bdd::BddCompare());
      BddVector::const_iterator it=difv.begin(),end=difv.end();
      vector<string> effFunctors;
      for (;end!=it;++it) {
        string fName=it->VarName();
        effFunctors.push_back(string(fName,0,fName.find("(")));
      }
      printit(effFunctors);
    }
    Bdd step=before*UnprimedToPrimed(after);
    map<string,pair<Bdd,BddVector> >::const_iterator abit=mActionBdds.begin();
    while ( abit!=mActionBdds.end() &&
            mgr.Zero() == step*abit->second.first ) ++abit;
    assert( abit!=mActionBdds.end() );
    *gpLog << sep() << abit->first << flush;
    mPlan.push_back(abit->first);
  }
  *gpLog << ".\n";
}

void BddTransPlan::Run()
{
  Init();
  assert(!mgr.WasOverflow());
  mgr.DoGC();
  ForwardPass();
  BackwardPass();
}

