// FluentIterator.cc -*- c++ -*- Hans-Peter Stoerr Thu Feb  4 1999

#include <string>
#include <vector>
#include <map>

#include "general.h"
#include "FluentIterator.h"
#include "Domain.h"
#include "pddldomain.h"
#include "misc/mystl.h"

FluentIterator::FluentIterator()
  : mFname(""), mArgs(), mpDom(NULL)
{
  NOOP;
}

FluentIterator::FluentIterator(Domain& theDom)
  : mFname(""), mArgs(), mpDom(&theDom)
{
  mFtypeIter=mpDom->dom.atomtypes.begin();
  mFtypeEnd=mpDom->dom.atomtypes.end();
  NextFluent(TRUE);
}

FluentIterator::~FluentIterator()
{
  NOOP;
}

bool    FluentIterator::operator==(const FluentIterator& b) const
{
  if (""==b.mFname && ""==mFname) return TRUE;
  return (b.mFname==mFname)&&(b.mArgs==mArgs);
}

bool    FluentIterator::operator!=(const FluentIterator& b) const
{
  return ! (*this == b);
}

bool    FluentIterator::NextFluent(bool firsttime)
{
  assert(mFtypeIter!=mFtypeEnd);
  mFname=""; mIsFirst=TRUE;
  if (!firsttime) mFtypeIter++;
  while( mFtypeIter!=mFtypeEnd && mpDom->IsConstantFname(mFtypeIter->first) )
    mFtypeIter++;
  if( mFtypeIter==mFtypeEnd ) return FALSE; // we are done.
  mFname=mFtypeIter->first;
  // now we initialize the iterator vectors
  mBeginIters=vector<StringVector::const_iterator>(); // erase the vectors
  mIters=mBeginIters;
  mEndIters=mBeginIters;
  mArgs=Atomtype();
  Atomtype::const_iterator it;
  for(it=mFtypeIter->second.begin(); mFtypeIter->second.end()!=it; ++it) {
    // *it is the type of the current argument.
    if (0==mpDom->mTypeConstantsMap[*it].size()) {
      *gpLog << "PDDL-Bug: can't have empty set of fluents for type " << *it
             << " used as argument.\n";
      abort();
    }
    mBeginIters.push_back(mpDom->mTypeConstantsMap[*it].begin());
    mIters.push_back(mpDom->mTypeConstantsMap[*it].begin());
    mEndIters.push_back(mpDom->mTypeConstantsMap[*it].end());
    mArgs.push_back(*mIters.back());
  }
  return TRUE;
}

FluentIterator& FluentIterator::operator++()
{
  assert(""!=mFname);           // we can't increment enditerator.
  mIsFirst=FALSE;
  if (""==mFname || 0==mIters.size()) {
    NextFluent();
    return *this;
  } // end reached.
  vector<StringVector::const_iterator>::const_iterator itb; // iter on mBeginIters
  vector<StringVector::const_iterator>::iterator iti; // iter on mIters
  vector<StringVector::const_iterator>::const_iterator ite; // iter on mEndIters
  Atomtype::iterator ita;       // iterator on mArgs
  itb=mBeginIters.end()-1, iti=mIters.end()-1, ite=mEndIters.end()-1; ita=mArgs.end()-1;
  ++(*iti);
  while ( *iti==*ite ) {        // while the corresponding arg is finished
    if ( mIters.begin()==iti ) { // first arg went through, so we are ready with mFname
      if (NextFluent()) return *this; // if there is a next fluent we return this
      mFname=""; return *this;  // we are done with all fluents.
    }
    *iti=*itb;                  // reset iterator
    *ita=**iti;                 // set arg to first object of corresponding type
    --itb; --iti; --ite; --ita; // goto next argument
    ++(*iti);
  }
  *ita=**iti;
  return *this;
}

string FluentIterator::operator *()
{
  return accumulatesep(mArgs.begin(), mArgs.end(), mFname+"(", string(","))+")";
}
