// LiteralQueue.cc -*- c++ -*- unknown Sun Feb 27 2000

#include "LiteralQueue.h"
#include <functional>
#include <algorithm>

// ================ LiteralQueue ================

// nothing to do at the moment.
LiteralQueue::  LiteralQueue(FNameT fun,unsigned arity) : mFName(fun), mArity(arity) {};
LiteralQueue::~LiteralQueue(){}

// ---------------- LQArg ----------------

// nothing to do at the moment.
LiteralQueue::LQArg::LQArg(){}
LiteralQueue::LQArg::~LQArg(){}

// ---------------- Insertion and Making Available ----------------

void  LiteralQueue::InsertCallback(const CLiteral& l,LQCallback* c)
{ // insert callback c for all literals l is an instantiation of.
  CHECK(mFName==l.mFName,"Huh? InsertCallback called with wrong literal!");
  CHECK(l.mArgs.size()==mArity,"InsertCallback called with wrong arity!");
  LQArg* insertPoint=&top;
  vector<ArgT>::const_iterator it=l.mArgs.begin();
  for (;l.mArgs.end()!=it;++it) {
    if (IsVariable(*it)) insertPoint=&insertPoint->mInstantiated[AnonVariable];
    else insertPoint=&insertPoint->mInstantiated[*it];
  }
  insertPoint->mClients.push_back(c);
}

void    LiteralQueue::LiteralAvailable(const CLiteral& l, bool removeCbs /*=false*/)
  // propagates c to all superliterals of c.
{
  LiteralAvailableRun(l,removeCbs,0,top);
}

void    LiteralQueue::LiteralAvailableRun(const CLiteral& l, bool removeCbs, unsigned narg, LQArg& mem)
{
  if (narg==mArity) { // we are arrived at the end. Call Callbacks.
    if (0!=mem.mClients.size()) {
      vector<LQCallback*>::const_iterator it=mem.mClients.begin(), end=mem.mClients.end();
      for (;end!=it; ++it) (*it)->LiteralAvailable(l);
      if (removeCbs) mem.mClients.erase(mem.mClients.begin(),mem.mClients.end());
    }
    return;
  }
  // now we treat current arg as uninstantiated.
  if (mem.mInstantiated.end()!=mem.mInstantiated.find(AnonVariable))
    LiteralAvailableRun(l,removeCbs,narg+1,mem.mInstantiated[AnonVariable]);
  // if arg is indeed uninstantiated we are done
  if (IsVariable(l.mArgs[narg])) return;
  if (mem.mInstantiated.end()!=mem.mInstantiated.find(l.mArgs[narg]))
    LiteralAvailableRun(l,removeCbs,narg+1,mem.mInstantiated[l.mArgs[narg]]);  
}

/*

#include <iostream.h>
#include "LiteralQueue.h"

LiteralQueue    TestQ("p",2);
ostream*        gpLog=&cout;

class   TestCallback: public LQCallback
{
public:
  TestCallback(FNameT n,ArgT a1,ArgT a2) : l((FNameT)n,(ArgT)a1,(ArgT)a2) 
  {TestQ.InsertCallback(l,this);}
  virtual void  LiteralAvailable(const CLiteral& ln)
  {cout << l << " receives available " << ln << "\n";}
private:
  CLiteral      l;
};

int main()
{
  CLiteral l1("p","A","B");
  CLiteral l2("p","A","C");
  CLiteral l3("p","B","B");
  TestCallback c1("p","A","B");
  TestCallback c2("p","?","?");
  TestCallback c3("p","A","?x");
  TestCallback c4("p","?y","B");
  cout << l1 << "\n";
  TestQ.LiteralAvailable(l1);
  cout << l2 << "\n";
  TestQ.LiteralAvailable(l2);
  cout << l3 << "\n";
  TestQ.LiteralAvailable(l3);
  return 0;
}

*/
