// Options.cc -*- c++ -*- Hans-Peter Stoerr Mon Feb  8 1999

#include <stdlib.h>
#include <assert.h>
#include <iostream>
#include <iterator.h>
#include <set>
#include <map>
#include <string>
using namespace std;
#include "general.h"
#include "Options.h"
#include "misc/mystl.h"
#include "misc/misc.h"

typedef map<string,int,less<string> >           StringIntMap;
typedef set<string,less<string> >               StringSet;
typedef map<string, string, less<string> >      StringStringMap;

static  struct  OptionGlobals {
  StringIntMap    mOptions;
  StringSet       mOptUsed;
  StringSet       mOptRegistered;
  StringStringMap mDescriptions;
  StringStringMap mDepends;
}       *sOptGlobals=NULL;

void    SetOption(const char* name,int value) { // set option
  if(!sOptGlobals->mOptRegistered.count(name)) {
    cerr << "Unknown option " << name << " given to SetOption.\n";
    return;
  }
  sOptGlobals->mOptions[name]=value;
  // set options the option depends on
  string depname=name;
  while (1==sOptGlobals->mDepends.count(depname)) {
    depname=sOptGlobals->mDepends[depname];
    if (1!=sOptGlobals->mOptions[depname]) sOptGlobals->mOptions[depname]=1;
  }
} 

void    SetOptions(const char* options)
  // set multiple options separated with , and assign values with =.
{
  StringSet opts(Split(options,","));
  StringSet::const_iterator it=opts.begin();
  for(; opts.end()!=it; ++it){
    pair<string,string> p=SplitPair(*it,"=");
    SetOption(p.first.c_str(),(""==p.second ? 1 : atoi(p.second.c_str())));
  }
}

void    PrintOptions() {        // print Optionset.
  StringIntMap::const_iterator it=sOptGlobals->mOptions.begin();
  *gpLog << "c Options:";
  for(; sOptGlobals->mOptions.end()!=it; it++)
    if(0!=it->second)
      *gpLog << " " << it->first << "=" << it->second;
  *gpLog << ".\n";
}

void    DescribeOptions(bool called) {     // print description of all registered options to cerr/cout
  ostream* out; if (called) out=&cout; else out=&cerr;
  *out << "Registered Options (with current values):\n";
  StringStringMap::const_iterator it=sOptGlobals->mDescriptions.begin();
  for(; sOptGlobals->mDescriptions.end()!=it; ++it){
    *out << it->first << "(" << sOptGlobals->mOptions[it->first] << "): " << it->second << "\n";
  }
}

Option::operator int() {        // query an option
  assert(sOptGlobals->mOptions.end()!=sOptGlobals->mOptions.find(name));
  sOptGlobals->mOptUsed.insert(name);
  return sOptGlobals->mOptions[name];
}

Option::Option(const char* thename, const char* description, int defaultvalue)
  : name(thename)
{
  if (NULL==sOptGlobals) sOptGlobals=new OptionGlobals;
  sOptGlobals->mOptions[name]=defaultvalue;
  sOptGlobals->mOptRegistered.insert(name);
  sOptGlobals->mDescriptions[name]=description;
}

DependentOption::DependentOption(const char* thename, const char* depends_on,
                                 const char* description, int defaultvalue)
  : Option(thename, description, defaultvalue)
{
  sOptGlobals->mDepends[thename]=depends_on;
}

static struct CheckUsedOptions
{
  CheckUsedOptions::~CheckUsedOptions()
    // check if there were options which were declared but not used.
  {
    // we only check this if the program was actually run, i.e. at least one option was queried.
    if(0==sOptGlobals->mOptUsed.size()) return; 
    StringSet   opt_not_used;
    StringIntMap::const_iterator it=sOptGlobals->mOptions.begin();
    for(; sOptGlobals->mOptions.end()!=it; it++)
      if(0==sOptGlobals->mOptUsed.count(it->first) &&
         (0==sOptGlobals->mDepends.count(it->first) || 0==sOptGlobals->mOptUsed.count(sOptGlobals->mDepends[it->first])) )
        {                       // :TODO: deeper nesting of dependencies
        opt_not_used.insert(it->first);
      }
    if(0!=opt_not_used.size()){
      cerr << "Internal warning: Option(s) ";
      copy(opt_not_used.begin(),opt_not_used.end(),
           ostream_iterator<string>(cerr," "));
      cerr << "never asked for.\n";
    }
    delete sOptGlobals;
  }
} sCheckThem;
  
