// Hey Emacs! This is -*- c++ -*- ! :-)
// Bool3.cc	-- Hans-Peter Stoerr Thu Jan 28 1999
// $Id: Bool3.cc,v 2.1 2000/02/14 16:20:27 stoerr Exp $

#include <assert.h>
#include <iostream.h>

#include "Bool3.h"

Bool3::Bool3(char c)
{
  switch(c) {
  case 'f': case 'F': case '0': case 0:
    mVal=F; break;
  case 't': case 'T': case '1': case 1:
    mVal=T; break;
  case 'u': case 'U':
    mVal=U; break;
  default:  assert(0);
  }
}

Bool3::Bool3(int i)
{
  if(i) mVal=T;
  else  mVal=F;
}

char Bool3::Char() const
{
  switch(mVal) {
  case F: return 'F';
  case U: return 'U';
  case T: return 'T';
  default: assert(0);
  }
}

Bool3 Bool3::operator -() const // logical negation
{
  switch(mVal) {
  case F: return T;
  case U: return *this;
  case T: return F;
  default: assert(0);
  }
}


Bool3 Bool3::operator +(const Bool3 b) const // logical or
{
  // truth table:
  // F: FUT
  // U: UUT
  // T: TTT
  if (T==mVal) return *this;
  if (T==b.mVal) return b;
  // T!=*this, T!=b
  if (b.mVal==U) return b;
  // b==F, *this== U or F
  return *this;
}

Bool3 Bool3::operator *(const Bool3 b) const // logical and
{
  // truth table:
  // F: FFF
  // U: FUU
  // T: FUT
  if (F==mVal) return *this;
  if (F==b.mVal) return b;
  // F!=*this, F!=b
  if (b.mVal==U) return b;
  // b==T, *this== U or T
  return *this;
}

Bool3 Bool3::operator <(const Bool3 b) const // logical implication.
{
  // truth table:
  // F: TTT
  // U: UUT
  // T: FUT
  if (T==b.mVal) return b;
  if (F==mVal) return T;
  // F!=*this, T!=b
  if (b.mVal==U) return b;
  // b==F, *this== U or T
  return *this;
}

Bool3 Bool3::operator ^(const Bool3 b) const // logical xor.
{
  // truth table:
  // F: FUT
  // U: UUU
  // T: TUF
  if (U==b.mVal) return b;
  if (U==mVal) return *this;
  // U!=*this, U!=b
  if (b.mVal==F) return *this;
  // b==T -> return negation of *this
  switch(mVal) {
  case F: return T;
  case T: return F;
  default: assert(0);
  }
}

Bool3 Bool3::Equiv(const Bool3 b) const // logical equivalence
{
  // truth table:
  // F: TUF
  // U: UUU
  // T: FUT
  if (U==b.mVal) return b;
  if (U==mVal) return *this;
  // U!=*this, U!=b
  if (b.mVal==T) return *this;
  // b==F -> return negation of *this
  switch(mVal) {
  case F: return T;
  case T: return F;
  default: assert(0);
  }
}

ostream& operator <<(ostream& s, const Bool3 b)
{
  s<<b.Char();
  return s;
}

istream& operator >>(istream& s, Bool3& b)
{
  char c;
  s>>c;
  b=Bool3(c);
  return s;
}

#if 1

static void printtruthtable(Bool3 (*op)(const Bool3, const Bool3))
{
  Bool3 a,b;
  a=false3;
  cout << a << ":" << op(a,false3) << op(a,undefined3) << op(a,true3) << "\n";
  a=undefined3;
  cout << a << ":" << op(a,false3) << op(a,undefined3) << op(a,true3) << "\n";
  a=true3;
  cout << a << ":" << op(a,false3) << op(a,undefined3) << op(a,true3) << "\n";
}

static Bool3 op(const Bool3 a,const Bool3 b)
{
  return a==b;
}

void doatest()
{
  Bool3 a,b,c,d(a);
  d=b;
  // Bool3 (Bool3::* op)(Bool3);
  // op=&Bool3::operator*;
  printtruthtable(op);
}

#endif
