/* Bool3.h	-- Hans-Peter Stoerr Thu Jan 28 1999
 * $Id: Bool3.h,v 2.1 2000/02/14 16:20:27 stoerr Exp $
 * class Bool3 for 3 valued logic
 */

#ifndef _BOOL3_H
#define _BOOL3_H

#include <assert.h>

class ostream;
class istream;

class Bool3
{
public:
  Bool3() { mVal=F; }
  Bool3(char c);
  Bool3(int i);                 // true if !=0, false if ==0

  // use equivalent compiler generated versions:
  // inline Bool3(const Bool3& b) { mVal=b.mVal; }
  // inline ~Bool3() {}
  // inline Bool3& operator =(const Bool3& b) { mVal=b.mVal; }

  inline bool   operator ==(const Bool3& b) const { return mVal==b.mVal; }

  Bool3 operator -() const; // logical negation

  Bool3 operator +(const Bool3 b) const; // logical or
  Bool3 operator *(const Bool3 b) const; // logical and
  Bool3 operator <(const Bool3 b) const; // logical implication.
  Bool3 operator ^(const Bool3 b) const; // logical xor.

  Bool3& operator +=(const Bool3 b) { *this=*this+b; return *this; }
  Bool3& operator *=(const Bool3 b) { *this=*this*b; return *this; }

  Bool3 Equiv(const Bool3 b) const; // logical equivalence
  
  char  Char() const;

  static Bool3 undefined3() { return U; }
  static Bool3 true3() { return T; }
  static Bool3 false3() { return F; }

  friend ostream& operator <<(ostream& s, const Bool3 b);
  friend istream& operator >>(istream& s, Bool3& b);
  
private:
  enum  Value {F, U, T};
  
  Bool3(Value v) { mVal=v; }
  Value mVal;
};

const   Bool3   true3=Bool3::true3();
const   Bool3   false3=Bool3::false3();
const   Bool3   undefined3=Bool3::undefined3();

#endif
