// Hey Emacs! This is -*- c++ -*- ! :-)
// pddldomain.cc	-- Hans-Peter Stoerr Thu Jan 14 1999

#include <iomanip.h>
#include <algo.h>

#include "general.h"
#include "pddldomain.h"
#include "misc/mystl.h"

extern  PddlDatabase    *yydatabase;
int yyparsefile(const char *name);
extern  int yymylineno;
extern  const char* yyfilename;

// ================================================================

PddlItem::PddlItem()
  : filename(yyfilename), linenumber(yymylineno)
{
  NOOP;
}

PddlItem::~PddlItem()
{
  NOOP;
}

// ================================================================

PddlDomain::PddlDomain(string *thename)
  : name(*thename)
{
  delete thename;
}

PddlDomain::~PddlDomain()
{
  NOOP;
};

void  PddlDomain::Extends(string *)
{
  UNIMPLEMENTED;
}

void  PddlDomain::Requires(StringSet *newreqs)
{
  copy(newreqs->begin(),newreqs->end(),
       inserter(requirements,requirements.begin()));
  delete newreqs;
}

void  PddlDomain::SetSubtypes(StringStringMap *newsubtypes)
{
  //   copy(newsubtypes->begin(),newsubtypes->end(),
  //        inserter(directsupertypes,directsupertypes.begin()));
  for(StringStringMap::iterator it=(newsubtypes->begin());
      newsubtypes->end()!=it; ++it) {
    directsupertypes.insert(*it); // insert subtype
    Atomtype    typeat; typeat.push_back(it->first);
    if (atomtypes.end()!=atomtypes.find(it->first)) {
      *gpLog << "PDDL bug: Predicate " << it->first << " used twice.\n" << flush;
      abort();
    }
    atomtypes[it->first]=typeat;      // insert pseudoatom for type.
  }
  delete newsubtypes;
}

void  PddlDomain::SetConstants(StringStringMap *newconst)
{
  copy(newconst->begin(),newconst->end(),
       inserter(constants,constants.begin()));
  delete newconst;
}

void  PddlDomain::SetAtomtypes(StringAtomtypeMap *newatt)
{
  StringAtomtypeMap::const_iterator it=newatt->begin(), end=newatt->end();
  for (;end!=it; ++it) {
    if (atomtypes.end()!=atomtypes.find(it->first)) {
      *gpLog << "PDDL bug: Predicate " << it->first << " used twice.\n" << flush;
      abort();
    }
    atomtypes[it->first]=it->second;
  }
  delete newatt;
}

void  PddlDomain::SetTimeless(TermSet *newtimeless)
{
  copy(newtimeless->begin(),newtimeless->end(),
       inserter(timeless,timeless.begin()));
  delete newtimeless;
}

void  PddlDomain::AddSafety(Term *s)
{
  safeties.insert(*s);
  delete s;
}

void  PddlDomain::AddConstraint(Term *s)
{
  constraints.insert(*s);
  delete s;
}

void  PddlDomain::AddAxiom(PddlAxiom *ax)
{
  axioms.insert(*ax);
  delete ax;
}

void  PddlDomain::AddAction(PddlAction *ac)
{
  actions[ac->name]=(*ac);
  delete ac;
}

void    PddlDomain::Printit(ostream& s) const
{
  s << "Domain " << name << " begin"
    << "\n  reqs:\t" << requirements
    << "\n  subtyp:\t" << directsupertypes
    << "\n  const:\t" << constants
    << "\n  atomtypes:\t" << atomtypes
    << "\n  safeties:\t" << safeties
    << "\n  timeless:\t" << timeless
    << "\n  axioms:\t" << axioms
    << "\n  constraints:\t" << constraints
    << "\n  actionset:\t" << actions
    << "\nDomain " << name << " end.\n\n";
}

// ================================================================

void    PddlAction::Printit(ostream& s) const
{
  s << "Action " << name << "\n";
  s << "\tparms:\t" << parameters << "\n";
  s << "\tvars:\t" << vars << "\n";
  s << "\tprecond:\t" << precondition << "\n";
  s << "\teffect:\t" << effect << "\n";
}

PddlAction::PddlAction()
{
  NOOP;
}

PddlAction::PddlAction(string *thename)
{
  name=*thename;
  delete thename;
}

PddlAction::~PddlAction()
{
  NOOP;
}

bool  operator<(const PddlAction& a, const PddlAction& b)
{
  return a.name<b.name;
}

void  PddlAction::SetParameters(StringStringPairList *parms)
{
  parameters=*parms;
  delete parms;
}

void  PddlAction::SetPrecondition(Term* pre)
{
  CHECK(precondition.IsEmpty(), "Huh? More than one precondition?");
  precondition=*pre;
  delete pre;
}

void  PddlAction::SetEffect(Term* eff)
{
  CHECK(effect.IsEmpty(), "Huh? More than one effect?");
  effect=*eff;
  delete eff;
}

void  PddlAction::SetPosEffect(TermSet* eff)
{
  TermSet negated_effects;
  for (TermSet::const_iterator it=eff->begin(); eff->end()!=it; ++it) {
    negated_effects.insert(MakeTerm(Term::Not,*it));
  }
  Term  the_effect=MakeTerm(Term::And,*eff);
  Term  add_precondition=MakeTerm(Term::And,negated_effects);
  if (effect.IsEmpty()) {         // effect is still empty
    effect=the_effect;
  } else {
    effect=MakeTerm(Term::And,effect,the_effect);
  }
  if (precondition.IsEmpty()) precondition=add_precondition;
  else precondition=MakeTerm(Term::And,precondition,add_precondition);
  delete eff;
}

void  PddlAction::SetNegEffect(TermSet* eff)
{
  TermSet negated_effects;
  for (TermSet::const_iterator it=eff->begin(); eff->end()!=it; ++it) {
    negated_effects.insert(MakeTerm(Term::Not,*it));
  }
  Term  the_effect=MakeTerm(Term::And,negated_effects);
  Term  add_precondition=MakeTerm(Term::And,*eff);
  if (effect.IsEmpty()) {         // effect is still empty
    effect=the_effect;
  } else {
    effect=MakeTerm(Term::And,effect,the_effect);
  }
  if (precondition.IsEmpty()) precondition=add_precondition;
  else precondition=MakeTerm(Term::And,precondition,add_precondition);
  delete eff;
}

void  PddlAction::SetVars(StringStringMap *v)
{
  vars=*v;
  delete v;
}

// ================================================================

bool  operator<(const PddlAxiom& a, const PddlAxiom& b)
{
  return a.implicand<b.implicand;
}
  
void    PddlAxiom::SetVars(StringStringMap *s)
{
  vars=*s;
  delete s;
}

void    PddlAxiom::SetContext(Term *c)
{
  context=*c; delete c;
}

void    PddlAxiom::SetImplicand(Term *i)
{
  implicand=*i;
  delete i;
}

void    PddlAxiom::Printit(ostream& s) const
{
  s << "Axiom " << implicand
    << "\n vars:\t" << vars
    << "\n context:\t" << context
    << "\n.\n";
}

// ================================================================

PddlProblem::PddlProblem(string *thename)
  : name(*thename)
{
  delete thename;
}

PddlProblem::~PddlProblem()
{
  NOOP;
}

void  PddlProblem::Requires(StringSet* newreqs)
{
  copy(newreqs->begin(),newreqs->end(),
       inserter(requirements,requirements.begin()));
  delete newreqs;  
}
  
void  PddlProblem::BasedOnSituation(const PddlSituation& sit)
{
  if(sit.domainname!=domainname) {
    cerr << gProgName << ": Problem " << name << " is based on situation "
         << sit.name << " with different domainname.\n";
    exit(2);
  }
  copy(sit.constants.begin(),sit.constants.end(),
       inserter(constants,constants.begin()));
  if (0!=init.size() && 0!=sit.init.size()) { // :TODO: is aborting right here?
    cerr << gProgName << ": Problem " << name << " is based on situation "
         << sit.name << " *and* has initial state!.\n";
    exit(2);
  }
  if (0!=sit.init.size()) init=sit.init;
}

void  PddlProblem::SetObjects(StringStringMap* newconst)
{
  copy(newconst->begin(),newconst->end(),inserter(constants,constants.begin()));
  delete newconst;  
}

void  PddlProblem::SetInit(TermSet* newinit)
{
  copy(newinit->begin(),newinit->end(),
       inserter(init,init.begin()));
  delete newinit;  
}

void  PddlProblem::SetGoal(Term* t)
{
  goal=*t;
  delete t;
}

void  PddlProblem::Printit(ostream& s) const
{
  s << "Problem " << name
    << "\n domain:\t" << domainname
    << "\n reqs:\t" << requirements
    << "\n const:\t" << constants
    << "\n init:\t" << init
    << "\n goal:\t" << goal
    << "\n serial:\t" << serial << "\tparallel:\t" << parallel
    << ".\n";
}

// ================================================================

PddlSituation::PddlSituation(string* thename) : name(*thename)
{
  delete thename;
}

PddlSituation::~PddlSituation()
{
  NOOP;
}

void  PddlSituation::SetDomainname(string *thedomainname)
{
  domainname=*thedomainname; delete thedomainname;
}

void  PddlSituation::SetConstants(StringStringMap *theconst)
{
  copy(theconst->begin(),theconst->end(),inserter(constants,constants.begin()));
  delete theconst;
}

void  PddlSituation::SetInit(TermSet *newinit)
{
  copy(newinit->begin(),newinit->end(), inserter(init,init.begin()));
  delete newinit;
}

void PddlSituation::Printit(ostream& s) const
{
  s << "Situation " << name
    << "\n domain:\t" << domainname
    << "\n const:\t" << constants
    << "\n init:\t" << init
    << ".\n";
}

// ================================================================

PddlDatabase::PddlDatabase()
{
  NOOP;
}

PddlDatabase::~PddlDatabase()
{
  map<string, PddlDomain*, less<string> >::iterator itd;
  map<string, PddlProblem*, less<string> >::iterator itp;
  map<string, PddlSituation*, less<string> >::iterator its;
  for(itd=domains.begin(); domains.end()!=itd; ++itd)
    delete itd->second;
  for(itp=problems.begin(); problems.end()!=itp; ++itp)
    delete itp->second;
  for(its=situations.begin(); situations.end()!=its; ++its)
    delete its->second;
}

const PddlDomain* PddlDatabase::Domain(string d) const
{
  if (""==d) {
    if (1==domains.size()) return domains.begin()->second;
    if (0==domains.size()) 
      cerr << gProgName << " : No Domain found.\n";
    else
      cerr << gProgName << " : Domain not specified but found several domains.\n";
    exit(2);
  }
  if (0==domains.count(d)) {
    cerr << gProgName << " : Request for undefined domain " << d << ".\n";
    exit(2);
  }
  return domains.find(d)->second;
}

const PddlProblem* PddlDatabase::Problem(string p) const
{
  if (""==p) {
    if (1==problems.size()) return problems.begin()->second;
    if (0==problems.size())
      cerr << gProgName << " : No Problem found.\n";
    else
      cerr << gProgName << " : No Problem specified but found several problems.\n";
    exit(2);
  }
  if (0==problems.count(p)) {
    cerr << gProgName << " : Request for undefined problem " << p << ".\n";
    exit(2);
  }
  return problems.find(p)->second;
}

const PddlProblem* PddlDatabase::Problem(const string d, const string p) const
{
  const PddlProblem* res=NULL;;
  if (""==d)
    return Problem(p);
  if (""!=p) {
    res=Problem(p);
    CHECK(d==res->domainname,"problem "+p+" has domain "+res->domainname+" different from "+d+".");
    return res;
  }
  // just d is specified, p=""
  map<string, PddlProblem*, less<string> >::const_iterator it=problems.begin();
  for(; problems.end()!=it; ++it) {
    if (d==it->second->domainname) {
      if(NULL!=res) {
        cerr << gProgName << ": Problem not specified for domain " + d + " but found several problems.";
        exit(2);
      }
      res=it->second;
    }
  }
  cerr << gProgName << ": No problem for domain " << d << " found.\n";
  exit(2);
}

const PddlSituation* PddlDatabase::Situation(string s) const
{
  if (0==situations.count(s)) {
    cerr << gProgName << " : Request for undefined situation " << s << ".\n";
    exit(2);
  }
  return situations.find(s)->second;
}

void    PddlDatabase::RegisterProblem(PddlProblem* p)
{
  problems[p->name]=p;
}

void    PddlDatabase::RegisterDomain(PddlDomain* d)
{
  domains[d->name]=d;
}

void    PddlDatabase::RegisterSituation(PddlSituation* s)
{
  situations[s->name]=s;
}

bool    PddlDatabase::ParseFile(const char* fname)
{
  yydatabase=this;
  int res=yyparsefile(fname);
  if(res) abort();
  return true;
}

bool    PddlDatabase::ParseFiles(const vector<string> files)
{
  vector<string>::const_iterator it=files.begin();
  for(; files.end()!=it; ++it)
    if(!ParseFile(it->c_str())) return FALSE;
  return TRUE;
}

// this is not a particularily clean solution for printing, but I don't care.
static ostream& operator<<(ostream& s,pair<const string,Printable*>p)
{
  s<<*p.second;
  return s;
}

void    PddlDatabase::ListProblems() const
  // print all found problems to stdout
{
  map<string, PddlProblem*, less<string> >::const_iterator it=problems.begin();
  for(; problems.end()!=it; ++it) {
    cout << " " << it->first;
  }
  cout << flush;
}

void    PddlDatabase::ListDomains() const
  // print all found domains to stdout
{
  map<string, PddlDomain*, less<string> >::const_iterator it=domains.begin();
  for(; domains.end()!=it; ++it) {
    cout << " " << it->first;
  }
  cout << flush;
}

void    PddlDatabase::ListProblemsWithDomains() const
  // print all found problems with their domains to stdout
{
  map<string, PddlProblem*, less<string> >::const_iterator it=problems.begin();
  for(; problems.end()!=it; ++it) {
    cout <<  "-d " << it->second->domainname
         << " -p " << it->first
         << " " << it->second->filename;
    if (it->second->filename!=Domain(it->second->domainname)->filename)
      cout << " " << Domain(it->second->domainname)->filename;
    cout << " # line " << setw(5) << it->second->linenumber << "\n";
  }
  cout << flush;
}

void    PddlDatabase::Printit(ostream& s) const
{
  s << "Pddl Database\n";
  if(domains.size()>0) {
    s << "================================================================\n"
      << "Domains:\n";
    copy(domains.begin(), domains.end(),
               ostream_iterator<pair<const string,Printable*> >
               (s,"----------------------------------------------------------------\n"));
  }
  if(problems.size()>0) {
    s << "================================================================\n"
      << "Problems:\n";
    copy(problems.begin(), problems.end(),
               ostream_iterator<pair<const string,Printable*> >
               (s,"----------------------------------------------------------------\n"));
  }
  if(situations.size()>0) {
    s << "================================================================\n"
      << "Situations:\n";
    copy(situations.begin(), situations.end(),
               ostream_iterator<pair<const string,Printable*> >
               (s,"----------------------------------------------------------------\n"));
  }
}
