/* Lexanalyser for PDDL                  -*- text -*- */
/* pddllex.l	-- Hans-Peter Stoerr Thu Jan  7 1999
 * $Id: pddllex.l,v 4.3 2003/12/04 07:23:37 stoerr Exp $
 */

%{
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <functional>
using namespace std;
#include "general.h"
#include "pddldomain.h"

// #define YYSTYPE GrammarRetvalT
#include "pddlgram.tab.h"

// #define YY_USER_ACTION { ECHO; }

void my_lex_fatal_error(char *);
#define YY_FATAL_ERROR(msg) my_lex_fatal_error( msg )

int yymylineno;
const char *yyfilename;

string *downcasep(const char* const p);

%}

%option yylineno noyywrap nodefault noreject caseless batch
%option perf-report
/* %option debug */

CHAR [a-zA-Z0-9_-]
STRING {CHAR}+
REQUIREMENT :{CHAR}+
VARNAME [?]{CHAR}+
WHITE [ \t\f\r]+
NL \n
COMMENT ;.*$

%%

"("		{return ('(');}
")"		{return (')');}
"="		{return ('=');}
"-"		{return ('-');}

":action"	{return (ACTION);}
":axiom"	{return (AXIOM);}
":constants"	{return (CONSTANTS);}
":constraint"	{return (CONSTRAINT);}
":context"	{return (CONTEXT);}
":domain"	{return (DDOMAIN);}
":extends"	{return (EXTENDS);}
":effect"	{return (EFFECT);}
":effect+"	{return (EFFECTPOS);}
":effect-"	{return (EFFECTNEG);}
":goal"		{return (GOAL);}
":init"		{return (INIT);}
":implies"	{return (IMPLIES);}
":length"	{return (LENGTH);}
":objects"	{return (CONSTANTS);}
":parallel"	{return (PARALLEL);}
":parameters"	{return (PARAMETERS);}
":precondition"	{return (PRECONDITION);}
":predicates"	{return (PREDICATES);}
":requirements"	{return (REQUIREMENTS);}
":serial"	{return (SERIAL);}
":situation"	{return (DSITUATION);}
":timeless"	{return (TIMELESS);}
":types"	{return (TYPES);}
":vars"		{return (VARS);}


"addendum"	{return (ADDENDUM);}
"and"		{return (AND);}
"define"	{return (DEFINE);}
"domain"	{return (DOMAINSTR);}
"exists"	{return (EXISTS);}
"forall"	{return (FORALL);}
"isgoal"	{return (ISGOAL);}
"imply"		{return (IMPLY);}
"isinitial"	{return (ISINITIAL);}
"limited"	{return (LIMITED);}
"not"		{return (NOT);}
"or"		{return (OR);}
"problem"	{return (PROBLEM);}
"situation"	{return (SITUATION);}
"when"		{return (WHEN);}
"either"	{return (EITHER);}
"fluent"	{return (FLUENT);}

{STRING}	{yylval.anstring=downcasep(yytext); return (NAME);}
{VARNAME}	{yylval.anstring=downcasep(yytext); return (VARNAME);}
{REQUIREMENT}	{yylval.anstring=downcasep(yytext); return (REQUIREMENT);}
{WHITE} ;
{COMMENT} {yymylineno++; /* if(0) goto find_rule; */ };
			/* goto find_rule eliminates annoying warning. */
{NL} {yymylineno++;};


%%

static void _muell()            // eliminate some warnings
{
  return;
  _muell();
  yy_flex_realloc(NULL,0);
  yyunput(0,NULL);
  yy_fatal_error(NULL);
}

string *downcasep(const char* const p)
{
  string *res=new string(p);
  transform(res->begin(),res->end(),res->begin(),ptr_fun<int,int>(tolower));
  return res;
}

void printerrorlocation() {
  const int printlen=56;
  char *s,*end,holdchar;
  char *pos=yytext;
  char *buf=yy_current_buffer->yy_ch_buf;
  int siz=yy_current_buffer->yy_n_chars;
  fprintf(stderr,"%s:%d: Scan error.\n",yyfilename,yylineno);
  if(!pos) return;
  if(yy_c_buf_p) {
    holdchar=*yy_c_buf_p;
    *yy_c_buf_p=yy_hold_char;
  }
  if(buf<pos){
    fprintf(stderr,"Leading context: ");
    s=pos-printlen;
    if(s<buf) s=buf;
    fprintf(stderr,"... ");
    for(;s<pos;s++) putc(*s,stderr);
    fprintf(stderr,"\n");
  }
  if(pos<buf+siz) {
    fprintf(stderr,"Trailing context: ");
    s=pos; end=buf+siz;
    if(end-s>printlen) end=s+printlen;
    for(;s<end;s++) putc(*s,stderr);
    fprintf(stderr," ...\n");
  }
  if(yy_c_buf_p) *yy_c_buf_p=holdchar;
}

void my_lex_fatal_error(char *msg) 
{
  fprintf(stderr,"Lexer Error: %s\n",msg);
  printerrorlocation();
  exit(YY_EXIT_FAILURE);
}


